import boto3
import os

class DataAnalysis:

    
    @staticmethod
    def calculate_sentiment(data_df):
        
        # Add tweet sentiment column to dataframe
        data_df['sentiment'] = data_df.apply(
            lambda row: DataAnalysis.get_sentiment(row['text'], 
                                                   row['lang']), 
            axis=1)
        
        return data_df
    
    
    @staticmethod
    def get_sentiment(text, language):
        # Define comprehend object
        comprehend = boto3.client(service_name='comprehend',
                                  region_name='us-east-1')
        # Return undefined if language is not recognized by comprehend
        if language not in ['hi', 'de', 'zh-TW', 'ko', 'pt', 'en', 
                            'it', 'fr', 'zh', 'es', 'ar', 'ja']:

            return 'undefined'
        
        else:
            # Get detected sentiment as lower case string
            return comprehend.detect_sentiment(
                Text=text, LanguageCode=language)['Sentiment'].lower()
        
    @staticmethod
    def get_top_n(tweet_df, column, top_n):
        sorted_df = tweet_df.sort_values(column, ascending=False)[:top_n]
        return sorted_df


import re
import datetime
import html


class TwitterObjectParsing:  
    # Parsing tweet object
    def process_tweet_object(self, tweet_json):       
        # Retweet 
        if 'retweeted_status' in tweet_json:
            return self.__parse_retweeted_tweet(tweet_json)
    
        # Quote  
        if tweet_json['is_quote_status']:
            return self.__parse_quoted_tweet(tweet_json)
        
        # Reply 
        if tweet_json['in_reply_to_status_id'] is not None:
            return self.__parse_reply_tweet(tweet_json)

        # Tweet 
        return self.__parse_tweet(tweet_json)
    
    def __extract_hashtags(self, status_text):
        """ Extract hashtags from a string """
        hashtags = re.findall(r"#\w+", status_text)
        return list(set([h.lower() for h in hashtags]))
    
    def __extract_mentions(self, status_text):
        """ Extract mentions from a string """
        mentions = re.findall(r"@\w+", status_text)
        return list(set([m.lower() for m in mentions]))
    
    def __remove_media_urls (self, status_text, media_url):
        """Remove Media URLs from text (photos, video, gif), 
        twitter adds media url by default""" 
        return status_text.replace(media_url, "")
    
    def __remove_appended_url(self, tweet_text):
        """Remove quoted status URL from tweet text 
        (Twitter adds quoted status URL  at the end of tweet text by defualt)"""
        return re.sub(r'(https?:\/\/t\.co/\w+)$', '', tweet_text)
    
    def __remove_status_urls(self, urls_list):
        # Remove status URL from URLs list(Twitter adds quoted tweet link to the quote text)
        return [u for u in urls_list
                  if 'https://twitter.com' not in u and 'status' not in u]
    
    def __parse_created_at_date(self, input_date):
        # Use re to get rid of the milliseconds.
        input_date = re.sub("\+\d+\s","", input_date)

        # Convert the string into datetime object
        output_date = datetime.datetime.strptime(input_date, "%a %b %d %H:%M:%S %Y")
        
        return output_date
    
    def __parse_tweet_coordinates(self, coordinates, coord_type):
        """ This function extracts the coordinates from the tweet object """

        # Check the coordinates format
        if coordinates['type'] == 'Point':
            # Take the value directly
            crds = coordinates['coordinates']

        elif coordinates['type'] == 'Polygon':
            # Take the average over the four points
            crds = np.array(coordinates['coordinates'][0]).mean(axis=0)
        
        else:
            crds = [None, None]

        # Check if they come from the field 'coordinates' or 'geo'
        if coord_type == 'coordinates':
            long, lat = crds[0], crds[1]

        elif coord_type == 'geo':
            long, lat = crds[1], crds[0]
        else:
            long, lat = None, None

        return long, lat
    
    def __parse_basic_tweet(self, tweet_json, tweet_type = 't', come_from = 'o',
        # Create a basic form of tweet and parse it
                          related_screen_name = None,
                          related_user_id_str = None,
                          related_tweet_id_str = None):

        tweet = dict()

        # Main details
        tweet['tweet_type'] = tweet_type
        tweet['created_at'] = self.__parse_created_at_date(
            tweet_json['created_at'])
        tweet['lang'] = tweet_json['lang']
        tweet['author_screen_name'] = tweet_json['user']['screen_name'].lower()
        tweet['author_name'] = tweet_json['user']['name']
        tweet['author_id_str'] = tweet_json['user']['id_str']
        tweet['tweet_id_str'] = tweet_json['id_str']
        #tweet['source'] = re.findall('>(.+)<', tweet_json['source'])[0]

        # Statistics
        tweet['retweet_count'] = tweet_json['retweet_count']
        tweet['favorite_count'] = tweet_json['favorite_count']

        # Relationship
        tweet['come_from'] = come_from
        tweet['related_screen_name'] = related_screen_name#.lower() if related_screen_name is not None else None
        tweet['related_user_id_str']  = related_user_id_str
        tweet['related_tweet_id_str'] = related_tweet_id_str

        # Text and entities (Will be filled according to the case)
        tweet['text'] = None 
        tweet['urls'] = None
        tweet['hashtags'] = None
        tweet['mentions'] = None
        tweet['photo'] = None
        tweet['video'] = None
        tweet['animated_gif'] = None
        
        return tweet
        
    def __update_tweet(self, tweet_json):
        # Update the tweet

        tweet = dict()
        # Original text
        tweet['text'] = html.unescape(tweet_json['full_text'])
        
        if 'media' in tweet_json['entities'].keys():     # Remove media url in tweet        
            # Remove the url from the tweet text
            tweet['text'] = self.__remove_media_urls(tweet['text'],
                            tweet_json['entities']['media'][0]['url'])
            
            # Add media URL to it's column
            tweet[tweet_json['extended_entities']['media'][0]['type']] = tweet_json[
                'extended_entities']['media'][0]['expanded_url']

        tweet['urls'] = [u['expanded_url'] for u in tweet_json['entities']['urls']]
        tweet['hashtags'] = self.__extract_hashtags(tweet_json['full_text'])
        tweet['mentions'] = self.__extract_mentions(tweet_json['full_text'])

        return tweet
    
    def __parse_tweet(self, tweet_json, tweet_type = 't', come_from = 'o',
                        related_screen_name = None,
                        related_user_id_str = None,
                        related_tweet_id_str = None):
        
        """ Parse the normal tweet"""
        tweet = self.__parse_basic_tweet(tweet_json, tweet_type, come_from,
                          related_screen_name, related_user_id_str, related_tweet_id_str)
        
        tweet.update(self.__update_tweet(tweet_json))
        return [tweet]
    
    def __parse_reply_tweet(self, tweet_json, come_from = 'o' , tweet_type = 'rp'):
        """ Parse Reply tweet (it's like normal but differ in some reply info) """
        return self.__parse_tweet(tweet_json, tweet_type = tweet_type , come_from = come_from,
                          related_screen_name = tweet_json['in_reply_to_screen_name'].lower(),
                          related_user_id_str = tweet_json['in_reply_to_user_id_str'],
                          related_tweet_id_str = tweet_json['in_reply_to_status_id_str'])
    
    
    def __parse_retweeted_tweet(self, tweet_json):         
        """ Parse retweet and retweeted tweet """

        # Parse the retweeted tweet
        if tweet_json['retweeted_status']['in_reply_to_status_id'] is not None: 
            # If retweeted tweet is reply
            retweeted_tweet = self.__parse_reply_tweet(
                tweet_json['retweeted_status'], come_from = 'rt')
        
        elif tweet_json['retweeted_status']['is_quote_status']:
            retweeted_tweet = self.__parse_quoted_tweet(
                tweet_json['retweeted_status'], come_from = 'rt')
            
        else:   # If retweeted tweet is a normal tweet (Retweet to Normal)
            retweeted_tweet = self.__parse_tweet(
                tweet_json['retweeted_status'], come_from = 'rt')

        # Parse the retweet, assign to it same info of retweeted tweet.
        tweet = self.__parse_basic_tweet(tweet_json, tweet_type = 'rt',
                            related_screen_name = retweeted_tweet[0]['author_screen_name'].lower(),
                            related_user_id_str = retweeted_tweet[0]['author_id_str'],
                            related_tweet_id_str = retweeted_tweet[0]['tweet_id_str'])
        
        # Media
        tweet['photo'] = retweeted_tweet[0]['photo']
        tweet['video'] = retweeted_tweet[0]['video']
        tweet['animated_gif'] = retweeted_tweet[0]['animated_gif']
        
        # Text text
        tweet['text'] = retweeted_tweet[0]['text']
        
        # Entities
        tweet['hashtags'] = retweeted_tweet[0]['hashtags']
        tweet['mentions'] = retweeted_tweet[0]['mentions']
        tweet['urls'] = retweeted_tweet[0]['urls']
        
        # Interactions
        tweet['retweet_count'] = retweeted_tweet[0]['retweet_count'] 
        tweet['favorite_count'] = retweeted_tweet[0]['favorite_count'] 
        
        return [tweet, *retweeted_tweet]
    
    def __parse_quoted_tweet(self, tweet_json, come_from='o'):
        """ Parse quoted tweet and the quote """

        # Parse the quoted tweet
        if 'quoted_status' not in tweet_json:
            # Quoted tweet is not avaliable
            return self.__parse_tweet(tweet_json, tweet_type='q', come_from=come_from)
        else:  
            # Quoted tweet is avaliable
            
            if tweet_json['quoted_status']['in_reply_to_status_id'] is not None:
                # If quoted tweet is reply tweet (quote to reply case)
                quoted_tweet = self.__parse_reply_tweet(
                    tweet_json['quoted_status'], come_from='q')[0]

            elif tweet_json['quoted_status']['is_quote_status']:
                # If the quoted tweet is a quote(quote to quote case)
                
                quoted_tweet = self.__parse_tweet(
                    tweet_json['quoted_status'], tweet_type='q', come_from='q')[0]

                quoted_tweet['urls'] = self.__remove_status_urls(quoted_tweet['urls'])
                quoted_tweet['text'] = self.__remove_appended_url(quoted_tweet['text'])
            else:  
                # The quoted tweet is normal tweet
                quoted_tweet =self.__parse_tweet(
                    tweet_json['quoted_status'], come_from='q')[0]

            # Parse the quote
            tweet = self.__parse_tweet(tweet_json, tweet_type='q', come_from=come_from,
                                related_screen_name=quoted_tweet['author_screen_name'],
                                related_user_id_str=quoted_tweet['author_id_str'],
                                related_tweet_id_str=quoted_tweet['tweet_id_str'])[0]

            # Remove root URL from enduser text
            tweet['text'] = self.__remove_appended_url(tweet['text'])

            tweet['hashtags'] = list(set(tweet['hashtags'] + quoted_tweet['hashtags']))
            tweet['mentions'] = list(set(tweet['mentions'] + quoted_tweet['mentions']))
            
            tweet['photo'] = quoted_tweet['photo']
            tweet['video'] = quoted_tweet['video']
            tweet['animated_gif'] = quoted_tweet['animated_gif']

            # Remove the quote tweet url (Auto added by twitter)
            tweet['urls'] = self.__remove_status_urls(tweet['urls'])
            tweet['urls'] = list(set(tweet['urls'] + quoted_tweet['urls']))

            return [tweet, quoted_tweet]




from aws_s3_resource.s3_bucket_error import NumberRandomCharsException, NumberCharsBucketNameException
from aws_s3_resource.s3_bucket import S3Bucket
from aws_s3_resource.s3_object import S3Object
import os
import pickle as pk



class Setup:
    available_actions = ['tweet_delete_events','direct_message_mark_read_events',
                         'direct_message_indicate_typing_events','retweet_to_user_events','retweet_by_user_events',
                         'tweet_create_events','reply_to_user_events','reply_by_user_events',
                         'mention_to_user_events','message_to_user_events','message_by_user_events','follow_to_user_events','follow_by_user_events',
                         'unfollow_by_user_events','block_by_user_events','unblock_by_user_events','mute_by_user_events',
                         'unmute_by_user_events','favorite_to_user_events','favorite_by_user_events']
    
    def __init__(self, storing_medium, events_folder = None, bucket_name = None):
        # Setup bucket on s3 or local folder according to storing_medium
        if storing_medium == 's3':
            if not S3Bucket.is_available(bucket_name):
                self.bucket_name = bucket_name
                self.create_bot_bucket()
                self.create_actions_logs()
        else:
            if not os.path.exists(events_folder):
                self.events_folder = events_folder
                self.create_events_folder()
                self.create_events_log_files()




    def create_bot_bucket(self):
        # Create a Bucket
        S3Bucket.create(self.bucket_name)
        
    def create_actions_logs(self):
        for action in Setup.available_actions:
            loggs = list()

            # Save empty list to each action log
            file_dict = {
                'file_name': f'{action}.pkl',
                'file_bytes': pk.dumps(loggs)
            }
            
            S3Object.upload(self.bucket_name, file_dict, with_random_prefix=False)

    def create_events_folder(self):
        os.mkdir(self.events_folder)

    def create_events_log_files(self):
        for action in Setup.available_actions:
            loggs = list()
            with open(os.path.join(self.events_folder, f'{action}.pkl'), 'wb') as f:
                pk.dump(loggs, f)



import os
import io
import pickle as pk
from abc import ABCMeta, abstractmethod
from .twitter_account import TwitterAccountProxy
import random 
import json
import threading
from datetime import datetime
from aws_s3_resource.s3_bucket_error import NumberRandomCharsException, NumberCharsBucketNameException
from aws_s3_resource.s3_bucket import S3Bucket
from aws_s3_resource.s3_object import S3Object
import time



class TwitterBot:
    def __init__(self,  bot_crediential_file,
        strategy_time_interval_range, action_time_interval_range,
        strategy_class, strategy_params): 
    
        self.bot_account = TwitterAccountProxy(bot_crediential_file)
        self.strategy_time_interval = random.randint(*strategy_time_interval_range)
        self.action_time_interval = random.randint(*action_time_interval_range)
        self.strategy_class = strategy_class
        self.strategy_params = strategy_params
        
    
    def go(self,bucket_name=None):
        strategy = self.strategy_class(**self.strategy_params)
        try:
            actions = strategy.start()
            for action in actions:
                # Do action
                self.bot_account.do(action)
                time.sleep(self.action_time_interval * 60)

        except Exception as e:
            object_key = 'errors.pkl'
            errors = list()
            if S3Object.is_available(bucket_name, object_key):
                logg_bytes = S3Object.download(bucket_name, object_key)
                with io.BytesIO(logg_bytes) as f:
                    errors = pk.load(f)

            errors.append({'time': str(datetime.now()), 'error_message': str(e)})

            data_bytes = pk.dumps(errors)

            file_dict = {
                    'file_name': object_key,
                    'file_bytes': data_bytes
                }

            S3Object.upload(bucket_name, file_dict, with_random_prefix=False)

        threading.Timer(self.strategy_time_interval * 3600 , self.go).start()

        


from abc import ABCMeta, abstractmethod


class Strategy(metaclass=ABCMeta):
    def start(self):
        # Data Collection
        collected_data = self.collect()
        
        # Data Analysis
        analysed_data = self.analyse(collected_data)
        
        # Behave
        return self.behave(analysed_data)
    
    @abstractmethod
    def collect(self):
        pass
    
    @abstractmethod
    def analyse(self, collected_data):
        pass
    
    @abstractmethod
    def behave(self, analysed_data):
        pass




from .twitter_object_parsing import TwitterObjectParsing
from .trends_data_collection import Trends
from .twitter_data_collection import QueryTweetsCollector, FavoritedTweetsCollector, UserTimelineTweetsCollector
import pandas as pd


class DataCollection:
    
    # Collect Data based on keyword
    @staticmethod
    def collect_keyword_tweets(keyword, number_of_tweets, filter_on_english):  
        collector = QueryTweetsCollector(
            DataCollection.query_search_params(keyword)
        )
        
        collector.collect(filter_on_english, number_of_tweets)
        return collector.tweets
        
    # Create Search params dictionary for query search
    @staticmethod
    def query_search_params(keyword):
        search_params = {
            'q': keyword,
            'tweet_mode': 'extended',
            'lang': 'en',
            'include_entities': True,
            'count': 100,  # 100 is the maximum per request (Twitter docs)
            'result_type': 'mixed',
        }
        
        return search_params
    
    # Collect Data favorited by specific user 
    @staticmethod
    def collect_favorited_tweets(user_screen_name, number_of_tweets):  
        collector = FavoritedTweetsCollector(
            DataCollection.favorited_tweet_params(user_screen_name)
        )
        
        collector.collect(number_of_tweets)
        return collector.tweets
    
    # Create Search params dictionary for query search
    @staticmethod
    def favorited_tweet_params(user_screen_name):
        search_params = {
            'id': user_screen_name,
            'tweet_mode': 'extended',
            'lang': 'en',
            'include_entities': True,
            'count': 200   # 200 is the maximum per request(Twitter docs)
        }
        
        return search_params

    # Collect Data form user timeline 
    @staticmethod
    def collect_timeline_tweets(user_id, number_of_tweets):  
        collector = UserTimelineTweetsCollector(
            DataCollection.timeline_tweets_params(user_id)
        )
        
        collector.collect(number_of_tweets)
        return collector.tweets
    
    # Create Search params dictionary for user timeline data
    @staticmethod
    def timeline_tweets_params(user_id):
        search_params = {
            'id': user_id,
            'tweet_mode': 'extended',
            'lang': 'en',
            'include_entities': True,
            'count': 200,   # 200 is the maximum per request (Twitter docs),
            'trim_user': True
        }
        
        return search_params
    
    @staticmethod
    def get_trend_data(number_of_places, filter_on_english):
        # Get trends
        tr = Trends()
        trends_df = tr.get_current_trends_df(filter_on_english, number_of_places=number_of_places)
        
        return trends_df
    
    # Convert tweets to data frame
    @staticmethod
    def tweets_list_to_dataframe(tweets):
        parser = TwitterObjectParsing() 
        parsed_tweets = list()
        
        for tw in tweets:
            # Parsing tweet object 
            parsed_tweets.extend(parser.process_tweet_object(tw._json))

        if len(tweets) > 0:
            tweets_df = pd.DataFrame(parsed_tweets)

            # Drop duplicates
            tweets_df.drop_duplicates('tweet_id_str', inplace=True)
        else:
            tweets_df = pd.DataFrame(columns=[
                'author_id_str', 'author_name', 'author_screen_name',
                'created_at', 'favorite_count', 'hashtags', 'lang',
                'location', 'mentions', 'related_screen_name',
                'related_tweet_id_str', 'related_user_id_str', 'retweet_count',
                'source', 'full_text', 'tweet_id_str', 'urls', 'video'])
            
        return tweets_df


class DataCollectionStrategies:
    @staticmethod
    def collect_trends_tweets(number_of_places, number_of_trends, 
        number_of_tweets, filter_on_english):

        trends_df = DataCollection.get_trend_data(number_of_places, filter_on_english)
        trends_df = trends_df.sort_values(by='tweet_volume', ascending=False)
        trends_df = trends_df[:number_of_trends]
        
        trends_tweets = dict()
        for topic in trends_df.trend_topic:
            trends_tweets[topic] = DataCollection.collect_keyword_tweets(topic, number_of_tweets, 
                filter_on_english)

            if trends_tweets[topic] == []:
                continue

            trends_tweets[topic] = DataCollection.tweets_list_to_dataframe(trends_tweets[topic])
        
        
        return trends_tweets


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="twitter_bot_tracking",
    version="0.0.8",
    author="sammer sallam",
    author_email="samersallam92@gmail.com",
    description="This library is to ease the operation of creating twitter bots and tracking it's ativities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Samer92/twitter-bot-tracking",
    packages=["twitter_bot_tracking"],
    install_requires=["aws-s3-resource==0.0.2","boto3==1.12.37","botocore==1.15.37","certifi==2020.4.5.1",
                      "chardet==3.0.4","docutils==0.15.2","idna==2.9","jmespath==0.9.5",
                      "langdetect==1.0.8","numpy==1.18.2","oauthlib==3.1.0","pandas==1.0.3",
                      "PySocks==1.7.1","python-dateutil==2.8.1","pytz==2019.3","requests==2.23.0",
                      "requests-oauthlib==1.3.0","s3transfer==0.3.3","six==1.14.0","tweepy==3.8.0","urllib3==1.25.8"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import tweepy
import json
from abc import abstractmethod, ABCMeta


class DataAPIInterface(metaclass=ABCMeta):
    """ Interface class to all data source api """
    @abstractmethod
    def get_current_usage(self):
        pass
    

class TwitterAPI(tweepy.API, DataAPIInterface):
    """ This class is to get a signed in API
        Also it has some extra functionalities required to know the state of the API
    """
    def __init__(self, token, wait_on_rate_limit=False, 
                 timeout=10, retry_count=3):
        """ create a signed in twitter API
            Inputs:
            token: a dictionary of four keys consumer_key, consumer_secret, access_token, 
                   and access_token_secret
        """
        
        self.token = token
        
        auth = tweepy.OAuthHandler(self.token['consumer_key'],
                                   self.token['consumer_secret'])
        auth.set_access_token(self.token['access_token'],
                              self.token['access_token_secret'])
        super().__init__(auth, wait_on_rate_limit=wait_on_rate_limit,
                             timeout=timeout, retry_count=retry_count)
    
    def get_current_usage(self):
        """ Get the limit of and how much is used from each required API endpoint
            You can append to this dictionary what is important for you
        """
        rate_limit = self.rate_limit_status()
        

        return             {
                'rate_limit_status': rate_limit['resources']['application']['/application/rate_limit_status'],
                'show_friendship': rate_limit['resources']['friendships']['/friendships/show'],
                'get_user': rate_limit['resources']['users']['/users/show/:id'],
                'get_followers_list': rate_limit['resources']['followers']['/followers/list'],
                'get_followers_ids': rate_limit['resources']['followers']['/followers/ids'],
                'get_friends_list': rate_limit['resources']['friends']['/friends/list'],
                'get_friends_ids': rate_limit['resources']['friends']['/friends/ids'],
                'search_api': rate_limit['resources']['search'],
                'user_timeline': rate_limit['resources']['statuses']['/statuses/user_timeline'],
                'user_favourited_statuses': rate_limit['resources']['favorites']['/favorites/list'],
                'trends_available': rate_limit['resources']['trends']['/trends/available'],
                'trends_place': rate_limit['resources']['trends']['/trends/place']
            
            }
    
    @staticmethod
    def is_rate_limit_error(e):
        return '429' in e.reason or '88' in e.reason
    
    @staticmethod
    def resource_was_not_found(e):
        return '34' in e.reason


class TwitterAPIsList:
    """ This class manages the available twitter apis 
        It is a singleton class where the signing in happens only one time
    """
    _shared_state = {}   # This is required to make the class a singleton class

    def __new__(cls, *args, **kwargs):
        """ This is required to make the class a singleton class """
        obj = super().__new__(cls, *args, **kwargs)
        obj.__dict__ = cls._shared_state
        return obj
    
    def __init__(self, tokens_file_path='twitter_tokens.json'):
        """ Create a list of all available twitter apis  """
        
        if len(TwitterAPIsList._shared_state):
            # No need to sign in again
            print('No need to sign in again')
            return
        
        # List of the available apis
        self.apis = []
        
        with open(tokens_file_path, 'r') as f:
            twitter_tokens = json.load(f)
            
            #for token in twitter_tokens:
            self.apis.append(TwitterAPI(twitter_tokens))
        
        self.number_of_apis = len(self.apis)
    
    def __getitem__(self, key):
        """ Get one of the available twitter API accounts """
        if key < 0 or key > self.number_of_apis-1:
            raise IndexError(f'The index should be in the range 0 to {self.number_of_apis-1}')
        return self.apis[key]
